/*
** ntk_vil_utils.cc
** Login : <nicolas.burrus@ensta.fr>
** Started on  Thu Jun 23 23:45:37 2005 Nicolas Burrus
** $Id$
**
** Copyright (C) 2005, 2006, 2007 Nicolas Burrus
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <ntk/vxl/vil_utils.h>
#include <ntk/utils/math.h>
#include <ntk/utils/iostream.h>
#include <ntk/stats/distributions.h>

#include <vcl_cmath.h>
#include <vil/vil_convert.h>
#include <vil/vil_save.h>
#include <vil/algo/vil_sobel_3x3.h>

#include <vil/algo/vil_fft.h>
#include <vil/vil_border.h>

namespace ntk {

  const vcl_vector<vil_point>& vil_neighb_c8()
  {
    static vcl_vector<vil_point> n_c8;
    static CriticalSection lock;
    if (n_c8.size() == 0)
    {
      lock.enter();
      if (n_c8.size() == 0)
      {
        n_c8.push_back(vil_point(1,0)); n_c8.push_back(vil_point(-1,0));
        n_c8.push_back(vil_point(1,1)); n_c8.push_back(vil_point(-1,-1));
        n_c8.push_back(vil_point(0,1)); n_c8.push_back(vil_point(0,-1));
        n_c8.push_back(vil_point(-1,1)); n_c8.push_back(vil_point(1,-1));
      }
      lock.exit();
    }
    return n_c8;
  }

  vil_structuring_element
  vil_disk(double r)
  {
    vil_structuring_element el;
    el.set_to_disk(r);
    return el;
  }

  const vil_structuring_element&
  vil_win_c8p()
  {
    static const int pi_[] = { -1, -1, -1,  0,  0,  0,  1,  1,  1 };
    static const int pj_[] = { -1,  0,  1, -1,  0,  1, -1,  0,  1 };
    static const vcl_vector<int> pi(pi_, pi_+9);
    static const vcl_vector<int> pj(pj_, pj_+9);
    static const vil_structuring_element el (pi, pj);
    return el;
  }

  const vil_structuring_element&
  vil_win_c4p()
  {
    static const int pi_[] = { -1,  0,  1,  0,  0, };
    static const int pj_[] = {  0,  0,  0,  1, -1, };
    static const vcl_vector<int> pi(pi_, pi_+5);
    static const vcl_vector<int> pj(pj_, pj_+5);
    static const vil_structuring_element el (pi, pj);
    return el;
  }

  bool vil_save_round(const vil_float_image& im, const char* filename)
  {
    vil_image_view<vxl_byte> tmp;
    vil_convert_round(im, tmp);
    return vil_save(tmp, filename);
  }

  bool vil_save_stretch(const vil_float_image& im, const char* filename)
  {
    vil_image_view<vxl_byte> tmp;
    vil_convert_stretch_range(im, tmp);
    return vil_save(tmp, filename);
  }

  bool vil_save_stretch(const vil_int_image& im, const char* filename)
  {
    vil_image_view<vxl_byte> tmp;
    vil_convert_stretch_range(im, tmp);
    return vil_save(tmp, filename);
  }

  bool vil_save_round(const vil_int_image& im, const char* filename)
  {
    vil_image_view<vxl_byte> tmp (im.ni(), im.nj());
    for_all_xy(tmp) tmp(x,y) = im(x,y)%256;
    return vil_save(tmp, filename);
  }

  void vil_invert(const vil_image_view<bool>& src, vil_image_view<bool>& dest)
  {
    dest.set_size(src.ni(), src.nj(), src.nplanes());
    for_all_xyp(src)
      dest(x,y,p) = !src(x,y,p);
  }

  void vil_sobel_intensity(const vil_byte_image& src, vil_byte_image& dest,
                           const vil_border<vil_byte_image>& border)
  {
    vil_border_accessor<vil_byte_image> src_accessor(src, border);
    vil_byte_image tmp (src.ni()+2, src.nj()+2);
    for_all_xy(tmp) tmp(x,y) = src_accessor(x-1,y-1);

    vil_float_image gi, gj;
    vil_sobel_3x3(tmp, gi, gj);

    dest.set_size(src.ni(), src.nj(), src.nplanes());
    for_all_xy(dest)
      dest(x,y) = vnl_math_rnd((vcl_abs(gi(x+1,y+1))+vcl_abs(gj(x+1,y+1))) / 2.);
  }

  vil_byte_image ideal_subsample_div4(vil_byte_image& im)
  {
    ntk_assert(im.ni()%8 == 0 && im.nj()%8 == 0, "Bad dimensions.");
    vil_image_view< vcl_complex<double> > cim (im.ni(), im.nj());
    for_all_xy(cim)
      cim(x,y) = vcl_complex<double>(im(x,y), 0);
    vil_fft_2d_fwd(cim);
    unsigned limit_1 = (im.nj()/8)+1;
    unsigned limit_2 = (im.nj()/8)*7-1;
    for (uint32 y = limit_1; y <= limit_2; ++y)
      for (uint32 x = 0; x < cim.ni(); ++x)
        cim(x,y) = vcl_complex<double>(0,0);

    limit_1 = (im.ni()/8)+1;
    limit_2 = (im.ni()/8)*7-1;
    for (uint32 y = 0; y < cim.nj(); ++y)
      for (uint32 x = limit_1; x <= limit_2; ++x)
        cim(x,y) = vcl_complex<double>(0,0);

    vil_fft_2d_bwd(cim);

    double max_value = 0;
    double min_value = FLT_MAX;
    for_all_xy(cim)
    {
      max_value = vcl_max(max_value, cim(x,y).real());
      min_value = vcl_min(min_value, cim(x,y).real());
    }

    vil_byte_image rim(im.ni(), im.nj());
    for_all_xy(rim)
    {
      rim(x,y) = (unsigned)((255.*(cim(x,y).real()-min_value))
                          /(max_value-min_value));
      ntk_assert(flt_eq(cim(x,y).imag(), 0), "Complex value.");
    }

    vil_byte_image result(im.ni()/4, im.nj()/4);
    for_all_xy(result)
    {
#if 0
      double tmp = 0;
      for (int dy = 0; dy < 4; ++dy)
      for (int dx = 0; dx < 4; ++dx)
        tmp += rim(x*4+dx, y*4+dy);
      tmp /= 16.;
      result(x,y) = unsigned(roundf(tmp));
#endif
      result(x,y) = rim(x*4, y*4);
    }
    return result;
  }

  vil_byte_image ideal_subsample_div2(vil_byte_image& im)
  {
    ntk_assert(im.ni()%4 == 0 && im.nj()%4 == 0, "Bad dimensions.");

    double initial_min_value = 255.0, initial_max_value = 0.0;
    for_all_xy(im) {
      initial_min_value = vcl_min(initial_min_value, (double)im(x,y));
      initial_max_value = vcl_max(initial_max_value, (double)im(x,y));
    }
    ntk_assert(initial_max_value >= initial_min_value, "Bad energy.");

    vil_image_view< vcl_complex<double> > cim (im.ni(), im.nj());
    for_all_xy(cim)
      cim(x,y) = vcl_complex<double>(im(x,y), 0);
    vil_fft_2d_fwd(cim);
    unsigned limit_1 = (im.nj()/4)+1;
    unsigned limit_2 = (im.nj()/4)*3-1;
    for (uint32 y = limit_1; y <= limit_2; ++y)
      for (uint32 x = 0; x < cim.ni(); ++x)
        cim(x,y) = vcl_complex<double>(0,0);

    limit_1 = (im.ni()/4)+1;
    limit_2 = (im.ni()/4)*3-1;
    for (uint32 y = 0; y < cim.nj(); ++y)
      for (uint32 x = limit_1; x <= limit_2; ++x)
        cim(x,y) = vcl_complex<double>(0,0);

    vil_fft_2d_bwd(cim);

    vil_byte_image tim(im.ni(), im.nj());
    for_all_xy(tim)
      tim(x,y) = (unsigned char)cim(x,y).real();

    double max_value = 0;
    double min_value = FLT_MAX;
    for_all_xy(cim)
    {
      max_value = vcl_max(max_value, cim(x,y).real());
      min_value = vcl_min(min_value, cim(x,y).real());
    }

    vil_byte_image rim(im.ni(), im.nj());
    for_all_xy(rim)
    {
      double v = (cim(x,y).real()-min_value) /(max_value-min_value);
      // v \in [0,1]
      rim(x,y) = vnl_math_rnd((initial_max_value-initial_min_value)*v + initial_min_value);
      ntk_assert(flt_eq(cim(x,y).imag(), 0), "Complex value.");
    }

    vil_byte_image result(im.ni()/2, im.nj()/2);
    for_all_xy(result)
    {
#if 0
      double tmp = 0;
      for (int dy = 0; dy < 2; ++dy)
      for (int dx = 0; dx < 2; ++dx)
        tmp += rim(x*2+dx, y*2+dy);
      tmp /= 4.;
#endif
      result(x,y) = rim(x*2, y*2);
    }
    return result;
  }

  vil_byte_image well_sampled_noise(unsigned ni, unsigned nj, vnl_random& rgen)
  {
    unsigned next_ni_pow_2 = (unsigned)pow(2, ceil(log(ni*2)/log(2)));
    unsigned next_nj_pow_2 = (unsigned)pow(2, ceil(log(nj*2)/log(2)));
    vil_byte_image im(next_ni_pow_2, next_nj_pow_2);
    for_all_xy(im)
      im(x,y) = rgen(256);
    im = ideal_subsample_div2(im);
    vil_byte_image output(ni, nj); // crop it
    for_all_xy(output) output(x,y) = im(x,y);
    return output;
  }

  void juce_to_vil(const Image& image, ntk::vil_byte_image& result)
  {
    int line_stride, pixel_stride;
    const uint8* raw_data =
        image.lockPixelDataReadOnly(0, 0,
                                    image.getWidth(), image.getHeight(),
                                        line_stride, pixel_stride);

    result.set_size(image.getWidth(), image.getHeight());
    for_all_xy(result)
      result(x,y) = raw_data[pixel_stride*x + line_stride*y];
  }

  Image* vil_to_juce(const ntk::vil_byte_image& vil_im)
  {
    ntk_assert(vil_im != 0, "Null pointer");
    Image* im = new Image(Image::RGB, vil_im.ni(), vil_im.nj(), false);

    int line_stride, pixel_stride;
    uint8* raw_data = im->lockPixelDataReadWrite(0, 0, vil_im.ni(), vil_im.nj(),
        line_stride, pixel_stride);
  // im->setPixelData (0, 0, vil_im.ni(), vil_im.nj(), vil_im.top_left_ptr(), vil_im.ni());
    for (unsigned j = 0; j < vil_im.nj(); ++j)
      for (unsigned i = 0; i < vil_im.ni(); ++i)
    {
      const uint8 value = vil_im(i,j);
      for (unsigned k = 0; k < 3; ++k)
        raw_data[k + pixel_stride*i + line_stride*j] = value;
    }
    im->releasePixelDataReadWrite (raw_data);
    return im;
  }

} // end of ntk
