//
// ac_selector.cc
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "ac_selector.h"
#include "ac_segmentor.h"
#include "ac_threshold_database.h"
#include "ac_learn_process.h"

#include <ntk/ntk_with_macros.h>

using namespace ntk;

void AcSelector ::
startLearningMode()
{
  m_is_learning = true;
}

void AcSelector ::
stopLearningMode()
{
  m_is_learning = false;
}

void AcSelector ::
initialize(AcSegmentor& segmentor)
{
  m_segmentor = &segmentor;

  if (!isLearning())
  {
    PropertySet properties;
    properties.setValue("name", segmentor.name());
    properties.setValue("width", (int)segmentor.getImageToSegment().nj());
    properties.setValue("height", (int)segmentor.getImageToSegment().ni());
    m_thresholds = thresholdDatabase()->getThresholds(properties);
    if (m_thresholds.size() == 0)
      ntk_throw_exception(T("Could not load thresholds."));
  }

  initializeImpl();
}

/*************************************************************************/
/*************************************************************************/

namespace {

  double region_mean(const vil_byte_image& im, const Region& r)
  {
    double mean = 0;
    for (uint32 i = 0; i < r.size(); ++i)
      mean += im(r[i].x, r[i].y);
    mean /= r.size();
    return mean;
  }

}

void AcSelectorMeanThreshold ::
initializeImpl()
{
  m_region_features.clear();
}

void AcSelectorMeanThreshold::
finalize()
{
}

void AcSelectorMeanThreshold ::
computeRegionFeatures(RegionFeatures* features, uint32 r)
{
  const Region& r_ref = segmentor().getRegion(r);
  features->mean = region_mean(segmentor().getImageToSegment(),
                               r_ref);
  features->size = r_ref.size();
}

void AcSelectorMeanThreshold ::
getDifference(uint32 r1, uint32 r2, double* distance, bool* is_meaningful)
{
  RegionFeatures& r1_features = m_region_features[r1];
  if (r1_features.isNull()) computeRegionFeatures(&r1_features, r1);

  RegionFeatures& r2_features = m_region_features[r2];
  if (r2_features.isNull()) computeRegionFeatures(&r2_features, r2);

  double mean_diff = vcl_abs(r1_features.mean-r2_features.mean);

  *distance = mean_diff;

  if (is_meaningful)
  {
    const uint32 category = 0; // no size dependent threshold.
    if (isLearning())
    {
      segmentor().learner().addDistance(category, *distance);
      *is_meaningful = false;
    }
    else
    {
      *is_meaningful = *distance > m_thresholds[category];
    }
  }
}

void AcSelectorMeanThreshold::
r2MergedIntoR1(uint32 r1, uint32 r2)
{
  ntk_assert(r1 < r2, "Bad order for regions.");
  RegionFeatures& r1_features = m_region_features[r1];
  vcl_map<uint32, RegionFeatures>::iterator r2_it = m_region_features.find(r2);
  const RegionFeatures& r2_features = r2_it->second;

  uint32 r1_size = segmentor().getRegion(r1).size();
  ntk_assert(r1_size == (r1_features.size + r2_features.size), "Regions not merged.");
  double new_mean = r1_features.mean * r1_features.size
                    + r2_features.mean * r2_features.size;
  new_mean /= r1_size;
  r1_features.mean = new_mean;
  r1_features.size = r1_size;

  m_region_features.erase(r2_it);
}

/*************************************************************************/
/*************************************************************************/

String AcSelectorMVF :: name() const
{
  String s ("AcSelectorMVFWith");
  if (m_use_mean) s << "Mean";
  if (m_use_deviation) s << "Dev";
  if (m_use_frontier) s << "Frontier";
  s << "V1";
  return s;
}

void AcSelectorMVF ::
segmentorInitialized()
{
  const vil_byte_image& raw_im = segmentor().getImageToSegment();
  m_max_possible_size = raw_im.ni()*raw_im.nj()/2+1;

  // fill edge features.
  const vil_image_view<int32>& label_image = segmentor().getLabelImage();

  // Note: some edge points might be counted twice:
  // 1  3a 3
  // 1a 2  3  e.g. in this example, 1a is kept with 3a and 3b
  // 1  3b 3
  for_all_xy(label_image)
  {

#define update_features_if_different(im, otherx, othery) \
    { \
      if (im.in_range(otherx,othery) \
          && im(otherx,othery) != current_id) \
      { \
        uint32 other_id = im(otherx,othery); \
        EdgeFeatures& f = m_edge_features[RegionEdge(other_id, current_id)]; \
        uint8 delta = vcl_abs(int16(int16(raw_im(x,y))-int16(raw_im(otherx,othery)))); \
        if (f.isNull()) \
        { \
          f.min_frontier_contrast = delta; \
          f.frontier_size = 1; \
        } \
        else \
        { \
          f.min_frontier_contrast = vcl_min(f.min_frontier_contrast, delta); \
          f.frontier_size += 1; \
        } \
      } \
    }

    int32 current_id = label_image(x,y);
    update_features_if_different(label_image, x+1, y);
    update_features_if_different(label_image, x,   y+1);
    update_features_if_different(label_image, x+1, y+1);
    update_features_if_different(label_image, x+1, y-1);
#undef update_features_if_different
  }
}

void AcSelectorMVF ::
initializeImpl()
{
  m_region_features.clear();
  m_edge_features.clear();
  
  m_grad_log_distrib.resize(256);
  
  // 1 to make sure that there is a least one element, so that we don't get
  // null probabilities.
  vcl_fill(stl_bounds(m_grad_log_distrib), 1);
  
  const vil_byte_image& im = segmentor().getImageToSegment();
  for_all_xy(im)
  {
    if (im.in_range(x+1,y))
    { m_grad_log_distrib[vcl_abs(int(im(x,y))-int(im(x+1,y)))] += 1; }

    if (im.in_range(x,y+1))
    { m_grad_log_distrib[vcl_abs(int(im(x,y))-int(im(x,y+1)))] += 1; }

    if (im.in_range(x+1,y+1))
    { m_grad_log_distrib[vcl_abs(int(im(x,y))-int(im(x+1,y+1)))] += 1; }

    if (im.in_range(x+1,y-1))
    { m_grad_log_distrib[vcl_abs(int(im(x,y))-int(im(x+1,y-1)))] += 1; }
  }

  // Normalization.
  double n = 0;
  foreach_idx(i, m_grad_log_distrib) n += m_grad_log_distrib[i];  
  
  foreach_idx(i, m_grad_log_distrib) m_grad_log_distrib[i] /= n;
  for (int i = int(m_grad_log_distrib.size()-2); i >= 0; --i)
    m_grad_log_distrib[i] += m_grad_log_distrib[i+1];
  
  foreach_idx(i, m_grad_log_distrib) m_grad_log_distrib[i] = log(m_grad_log_distrib[i]);
}

/**
 * Check segmentation consistency.
 */
void AcSelectorMVF::
finalize()
{
  const vil_byte_image& raw_im = segmentor().getImageToSegment();

  // Check that regions features are accurate.
  foreach_const_it(region_it, m_region_features, cat2(vcl_map<uint32, RegionFeatures>))
  {
    const Region& r_ref = segmentor().getRegion(region_it->first);

    vcl_vector<uint32> real_distrib (256,0);
    foreach_idx(i, r_ref) real_distrib[raw_im(r_ref[i].x, r_ref[i].y)] += 1;

    const vcl_vector<uint32>& distrib = region_it->second.gray_distribution;

    ntk_assert(region_it->second.size == r_ref.size(), "Invalid size!");
    foreach_idx(i, real_distrib)
    {
      ntk_assert(real_distrib[i] == distrib[i], "Invalid distribution!");
    }
  }

  // Check that edge features are accurate.
  vcl_map<RegionEdge, EdgeFeatures>::const_iterator edge_it;
  for (edge_it = m_edge_features.begin(); edge_it != m_edge_features.end(); ++edge_it)
  {
    const Region& r1_ref = segmentor().getRegion(edge_it->first.r1());
    const Region& r2_ref = segmentor().getRegion(edge_it->first.r2());
    vil_image_view<int32> im (raw_im.ni(), raw_im.nj());
    im.fill(-1);
    for (size_t i = 0; i < r1_ref.size(); ++i) im(r1_ref[i].x, r1_ref[i].y) = r1_ref.id();
    for (size_t i = 0; i < r2_ref.size(); ++i) im(r2_ref[i].x, r2_ref[i].y) = r2_ref.id();

    uint32 frontier_size = 0;
    uint8 min_contrast = 255;
    for_all_xy(im)
    {
#define update_if_different(im, otherx, othery) \
      if (im.in_range(otherx,othery) \
          && im(otherx,othery) != current_id \
          && im(otherx,othery) != -1) \
      { \
        int other_id = im(otherx,othery); \
        frontier_size += 1; \
        min_contrast = vcl_min(min_contrast, \
                               (uint8)vcl_abs(int(raw_im(x,y))-int(raw_im(otherx,othery)))); \
      }

      int current_id = im(x,y);
      if (current_id == -1) continue;
      update_if_different(im, x+1, y);
      update_if_different(im, x,   y+1);
      update_if_different(im, x+1, y+1);
      update_if_different(im, x+1, y-1);
#undef update_if_different
    }

    ntk_assert(frontier_size == edge_it->second.frontier_size,
               "Frontier size should be equivalent!");
    ntk_assert(min_contrast == edge_it->second.min_frontier_contrast,
               "Frontier contrast should be equivalent!");
  }
}

void AcSelectorMVF ::
computeRegionFeatures(RegionFeatures* features, uint32 r)
{
  const Region& r_ref = segmentor().getRegion(r);
  const vil_byte_image& im = segmentor().getImageToSegment();
  features->gray_distribution.resize(256,0);
  for (uint32 i = 0; i < r_ref.size(); ++i)
    features->gray_distribution[im(r_ref[i].x, r_ref[i].y)] += 1;
  double mean, dev;
  distrib_moments(features->gray_distribution, mean, dev);
  features->dev = dev;
  features->size = r_ref.size();
}

void AcSelectorMVF ::
getDifference(uint32 r1, uint32 r2, double* distance, bool* is_meaningful)
{
  RegionFeatures& r1_features = m_region_features[r1];
  if (r1_features.isNull()) computeRegionFeatures(&r1_features, r1);

  RegionFeatures& r2_features = m_region_features[r2];
  if (r2_features.isNull()) computeRegionFeatures(&r2_features, r2);

  uint32 min_size = vcl_min(r1_features.size, r2_features.size);

  double mann_pfa = 0;
  if (m_use_mean)
  {
    mann_pfa = compute_mann_whitney_log_pfa(r1_features.gray_distribution,
                                            r2_features.gray_distribution,
                                            r1_features.size,
                                            r2_features.size);
  }

  double dev_pfa = 0;
  // estimations are not valid enough for regions under 30 pixels.
  if (m_use_deviation && min_size > 30)
  {
    vcl_vector<uint32> r12_distrib(256,0);
    foreach_idx(i, r12_distrib)
      r12_distrib[i] = r1_features.gray_distribution[i] + r2_features.gray_distribution[i];
    double gmean=0, gdev=0, gmu4=0;
    distrib_moments(r12_distrib, gmean, gdev, gmu4);
    double pfa1 = deviation_log_pfa(r1_features.dev, r1_features.size, gdev, gmu4);
    double pfa2 = deviation_log_pfa(r2_features.dev, r2_features.size, gdev, gmu4);
    dev_pfa = pfa1 + pfa2;
  }

  double frontier_pfa = 0;
  if (m_use_frontier)
  {
    const EdgeFeatures& f = m_edge_features[RegionEdge(r1,r2)];
    ntk_assert(!f.isNull(), "Edge feature should exist.");
    double n = 0.5 * f.frontier_size; // only almost independent pixels.
    frontier_pfa = n * m_grad_log_distrib[f.min_frontier_contrast];
  }

  double pfa = mann_pfa + dev_pfa + frontier_pfa;
  *distance = -pfa;

  if (is_meaningful)
  {
    const uint32 category = getCoupleCategory(r1_features, r2_features);

    if (isLearning())
    {
      segmentor().learner().addDistance(category, *distance);
      *is_meaningful = false;
    }
    else
    {
      *is_meaningful = *distance > m_thresholds[category];
    }
  }
}

void AcSelectorMVF::
r2MergedIntoR1(uint32 r1, uint32 r2)
{
  ntk_assert(r1 < r2, "Bad order for regions.");
  RegionFeatures& r1_features = m_region_features[r1];
  vcl_map<uint32, RegionFeatures>::iterator r2_it = m_region_features.find(r2);
  const RegionFeatures& r2_features = r2_it->second;

  // region features
  {
    uint32 r1_size = segmentor().getRegion(r1).size();
    ntk_assert(r1_size == (r1_features.size + r2_features.size), "Regions not merged.");

    for (uint32 i = 0; i < 256; ++i)
      r1_features.gray_distribution[i] += r2_features.gray_distribution[i];

    double mean = 0;
    distrib_moments(r1_features.gray_distribution, mean, r1_features.dev);
    r1_features.size = r1_size;
  }

  // edges features
  {
    const Region& r1_ref = segmentor().getRegion(r1);
    const Region& r2_ref = segmentor().getRegion(r2);

    Region::neighbors_const_iterator it;
    for (it = r2_ref.neighborsBegin(); it != r2_ref.neighborsEnd(); ++it)
    {
      if (*it == r1) continue;

      vcl_map<RegionEdge, EdgeFeatures>::iterator r2x_it;
      r2x_it = m_edge_features.find(RegionEdge(r2, *it));
      ntk_assert(r2x_it != m_edge_features.end(), "Feature should be here!");
      const EdgeFeatures& r2x_features = r2x_it->second;

      vcl_map<RegionEdge, EdgeFeatures>::iterator r1x_it;
      EdgeFeatures& r1x_features = m_edge_features[RegionEdge(r1, *it)];

      if (r1x_features.isNull())
      {
        r1x_features = r2x_features;
      }
      else
      {
        r1x_features.min_frontier_contrast =
            vcl_min(r1x_features.min_frontier_contrast,
                    r2x_features.min_frontier_contrast);
        r1x_features.frontier_size += r2x_features.frontier_size;
      }

      m_edge_features.erase(r2x_it);
    }
    m_edge_features.erase(RegionEdge(r1,r2));
  }

  m_region_features.erase(r2_it);
}

uint32 AcSelectorMVF ::
getCoupleCategory(const RegionFeatures& r1_features,
                  const RegionFeatures& r2_features)
{
  double size = vcl_min(r1_features.size, r2_features.size);
  uint32 max_index;
  if (isLearning())
    max_index = segmentor().learner().nbCategories();
  else
    max_index = m_thresholds.size()-1;

  const double d = 2.0; // controls curve's shape.
  double k = max_index-1;
  double idx = (k/d)*log(exp(0.0) + (size-1)*(exp(d)-exp(0.0))/m_max_possible_size);
  uint32 uidx = (unsigned)(idx); // floorf
  return uidx;
}
