//
// ac_selector.h
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef ACSEGMENTOR_ac_selector_H
# define ACSEGMENTOR_ac_selector_H

# include <ntk/core.h>

# include "region.h"

# include <vcl_map.h>
# include <vcl_utility.h>
# include <vcl_vector.h>

class AcSegmentor;
class AcLearner;
class AcThresholdDatabase; ntk_ptr_typedefs(AcThresholdDatabase);

class AcSelector
{
  public:
    AcSelector(const AcThresholdDatabasePtr& threshold_database)
    : m_threshold_database(threshold_database),
      m_segmentor(0), 
      m_is_learning(false)
    {}

    virtual ~AcSelector() {}

    virtual String name() const = 0;
    
  public:
    virtual void startLearningMode();
    virtual void stopLearningMode();
    bool isLearning() const { return m_is_learning; }
    
    void initialize(AcSegmentor& segmentor);    
    virtual void initializeImpl() = 0;
    virtual void segmentorInitialized() {}
    
    virtual void finalize() {};

  public:
    AcSegmentor& segmentor() { return *m_segmentor; }
    const AcSegmentor& segmentor() const { return *m_segmentor; }
    
    AcThresholdDatabasePtr thresholdDatabase() const
    { return m_threshold_database; }
    
  public:
    virtual void getDifference(uint32 r1, uint32 r2, double* score, bool* is_meaningful = 0) = 0;

  public:
    virtual void r2MergedIntoR1(uint32 r1, uint32 r2) = 0;
    
  protected:
    AcThresholdDatabasePtr m_threshold_database;
    vcl_vector<double> m_thresholds;
    AcSegmentor* m_segmentor;
    bool m_is_learning;
};
ntk_ptr_typedefs(AcSelector);

class AcSelectorMeanThreshold : public AcSelector
{
  public:
    struct RegionFeatures 
    {
      RegionFeatures() { mean = -1; size = 0; }
      bool isNull() const { return mean < 0; }
      double mean;
      uint32 size;
    };  
  
  public:
    AcSelectorMeanThreshold(const AcThresholdDatabasePtr& threshold_database)
      : AcSelector(threshold_database)
    {}
      
    virtual ~AcSelectorMeanThreshold() {}
    
    virtual String name() const { return String("AcSelectorMeanThresholdV1"); }
    
  public:
    virtual void initializeImpl();
    virtual void finalize();
      
    virtual void getDifference(uint32 r1, uint32 r2, double* distance, bool* is_meaningful);
    virtual void r2MergedIntoR1(uint32 r1, uint32 r2);

  private:
    void computeRegionFeatures(RegionFeatures* features, uint32 r);
    
  private:
    vcl_map<uint32, RegionFeatures> m_region_features;
};

class AcSelectorMVF : public AcSelector
{
  public:
    struct RegionFeatures 
    {
      RegionFeatures() { size = 0; dev = 0; }
      bool isNull() const { return size == 0; }
      vcl_vector<uint32> gray_distribution;
      uint32 size;
      double dev;
    };
  
    struct EdgeFeatures
    {
      EdgeFeatures() : min_frontier_contrast(0), frontier_size(0)
      {}
      
      bool isNull() const { return frontier_size == 0; }
      
      uint8 min_frontier_contrast;
      uint32 frontier_size;
    };
    
  public:
    AcSelectorMVF(const AcThresholdDatabasePtr& threshold_database)
      : AcSelector(threshold_database),
        m_max_possible_size(0),
        m_use_mean(true), 
        m_use_deviation(true),
        m_use_frontier(true) 
    {}
      
    virtual ~AcSelectorMVF() {}
    
    virtual String name() const; 
    
  public:
    void setUseMean(bool use_it) { m_use_mean = use_it; }
    void setUseDeviation(bool use_it) { m_use_deviation = use_it; }
    void setUseFrontier(bool use_it) { m_use_frontier = use_it; }    
        
  public:
    virtual void initializeImpl();
    virtual void segmentorInitialized();
    virtual void finalize();
      
    virtual void getDifference(uint32 r1, uint32 r2, double* distance, bool* is_meaningful);
    virtual void r2MergedIntoR1(uint32 r1, uint32 r2);

  private:
    void computeRegionFeatures(RegionFeatures* features, uint32 r);
    uint32 getCoupleCategory(const RegionFeatures& r1_features,
                           const RegionFeatures& r2_features);
    
  private:
    vcl_map<uint32, RegionFeatures> m_region_features;
    vcl_map<RegionEdge, EdgeFeatures> m_edge_features;
    uint32 m_max_possible_size;
    bool m_use_mean;
    bool m_use_deviation;
    bool m_use_frontier;
    vcl_vector<double> m_grad_log_distrib;
};

#endif // ndef ACSEGMENTOR_ac_selector_H
