//
// ac_threshold_database.cc
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "ac_threshold_database.h"
#include "ac_threshold_embedded_database.inc"

#include <ntk/ntk_with_macros.h>

class ThresholdsExtractor
{
  public:
    virtual ~ThresholdsExtractor() {}

  public:
    virtual String name() const = 0;
    virtual void loadFromXmlElement(const XmlElement& section) = 0;
    virtual void fillXmlElement(XmlElement& e) const = 0;

  public:
    virtual vcl_vector<double> getThresholds(const PropertySet& properties) = 0;
};

namespace {

  class SizeDependentThresholdsExtractor : public ThresholdsExtractor
  {
    public:
      SizeDependentThresholdsExtractor() {}

      virtual ~SizeDependentThresholdsExtractor() {}

    public:
      virtual String name() const { return T("SizeDependent"); };

      virtual void loadFromXmlElement(const XmlElement& section)
      {
        forEachXmlChildElement (section, e)
        {
          if (e->hasTagName (T("values")))
          {
            ntk_ensure(e->hasAttribute(T("width")), "No width for values.");
            ntk_ensure(e->hasAttribute(T("height")), "No height for values.");
            uint32 width = e->getIntAttribute(T("width"));
            uint32 height = e->getIntAttribute(T("height"));
            vcl_vector<double>& v = m_thresholds[vcl_make_pair(width, height)];
            String raw_values = e->getAllSubText();
            StringArray values_array; values_array.addTokens(raw_values, false);
            v.resize(values_array.size());
            for (int i = 0; i < values_array.size(); ++i)
              v[i] = values_array[i].getDoubleValue();
            ntk_dbg(1) << "  => " << (int)v.size() << " values loaded for size ("
                << width << ", " << height << ")";
          }
        }
      }

      virtual void fillXmlElement(XmlElement& e) const
      {
        threshold_map_type::const_iterator it;
        for (it = m_thresholds.begin(); it != m_thresholds.end(); ++it)
        {
          XmlElement* child = new XmlElement(T("values"));
          child->setAttribute(T("width"), (int)it->first.first);
          child->setAttribute(T("height"), (int)it->first.second);

          ntk_ensure(it->second.size() > 0, "Empty set of values.");
          String values;
          values << it->second[0];
          for (uint32 i = 1; i < it->second.size(); ++i)
            values << " " << it->second[i];

          XmlElement* text = XmlElement::createTextElement(values);
          child->addChildElement(text);
          e.addChildElement(child);
        }
      }

    public:
      virtual vcl_vector<double> getThresholds(const PropertySet& properties)
      {
        ntk_ensure(properties.containsKey(T("width")), "Width must be specified!");
        ntk_ensure(properties.containsKey(T("height")), "Height must be specified!");
        uint32 width = properties.getIntValue(T("width"));
        uint32 height = properties.getIntValue(T("height"));
        if (!ntk::has_key(m_thresholds, vcl_make_pair(width, height)))
          ntk_throw_exception(T("No thresholds were learnt for ")
                              << width << "x" << height << T(" images."));
        return m_thresholds[vcl_make_pair(width, height)];
      }

    private:
      typedef vcl_map<std::pair<uint32,uint32>, vcl_vector<double> > threshold_map_type;
      threshold_map_type m_thresholds;
  };

}

AcThresholdDatabase :: AcThresholdDatabase()
{
  String s (default_database);
  XmlDocument doc (s);
  XmlElement* main_element = doc.getDocumentElement();
  if (main_element == 0)
  {
    String error = doc.getLastParseError();
    ntk_throw_exception(error.toUTF8());
  }
  loadFromXmlElement(*main_element);
}

void
AcThresholdDatabase :: loadFromXmlElement(const XmlElement& main_element)
{
  forEachXmlChildElement (main_element, e)
  {
    if (!e->hasTagName(T("thresholds"))) continue;

    ntk_ensure(e->hasAttribute(T("name")), "No name attribute.");
    ntk_ensure(e->hasAttribute(T("loader")), "No name attribute.");
    String name = e->getStringAttribute(T("name"));
    String loader = e->getStringAttribute(T("loader"));

    if (loader == T("SizeDependent"))
    {
      ntk_dbg(1) << "Loading thresholds for: `" << name << "'";
      ThresholdsExtractorPtr& extractor = m_thresholds[name];
      if (!extractor || extractor->name() != loader)
        extractor.reset(new SizeDependentThresholdsExtractor());
      extractor->loadFromXmlElement(*e);
    }
    else
    {
      ntk_throw_exception("Loader does not exist.");
    }
  }
}

void
AcThresholdDatabase :: loadFromFile(const char* filename)
{
  File f (filename);
  XmlDocument doc(f);
  XmlElement* main_element = doc.getDocumentElement();
  if (main_element == 0)
  {
    String error = doc.getLastParseError();
    ntk_throw_exception(error.toUTF8());
  }
  loadFromXmlElement(*main_element);
}

void
AcThresholdDatabase :: saveToFile(const char* filename) const
{
  XmlElement e = createXmlElement();
  e.writeToFile(File(filename), String());
}

XmlElement
AcThresholdDatabase :: createXmlElement() const
{
  XmlElement root (T("root"));
  foreach_const_it(it, m_thresholds, threshold_map_type)
  {
    XmlElement* e = new XmlElement(T("thresholds"));
    e->setAttribute(T("name"), it->first);
    e->setAttribute(T("loader"), it->second->name());
    it->second->fillXmlElement(*e);
    root.addChildElement(e);
  }
  return root;
}

vcl_vector<double>
AcThresholdDatabase :: getThresholds(const PropertySet& properties)
{
  String name = properties.getValue("name");

  threshold_map_type::const_iterator it = m_thresholds.find(name);

  if (it == m_thresholds.end())
    ntk_throw_exception(T("No thresholds were learnt for these parameters `") << name << T("'"));

  return it->second->getThresholds(properties);
}
