//
// ac_properties.h
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef ACSEGMENTOR_gui_ac_properties_H
# define ACSEGMENTOR_gui_ac_properties_H

# include <ntk/ntk.h>

// Taken from jucer.
class FilePropertyComponent  : public PropertyComponent,
      public FilenameComponentListener
{
  public:
    //==============================================================================
    FilePropertyComponent(const String& name,
                          const bool isDirectory,
                          const bool allowEditingOfFilename,
                          const bool isForSave,
                          const String& fileBrowserWildcard = JUCE_T("*"))
        : PropertyComponent(name)
    {
      addAndMakeVisible(filenameComp = new FilenameComponent(name,
                                                             File::nonexistent,
                                                             allowEditingOfFilename,
                                                             isDirectory,
                                                             isForSave,
                                                             fileBrowserWildcard,
                                                             String::empty,
                                                             String::empty));

      filenameComp->addListener(this);
    }

    ~FilePropertyComponent()
    {
      deleteAllChildren();
    }

    //==============================================================================
    juce_UseDebuggingNewOperator

    //==============================================================================
    virtual void setFile(const File& newFile) = 0;
    virtual const File getFile() const = 0;

    //==============================================================================
    void refresh()
    {
      filenameComp->setCurrentFile(getFile(), false);
      filenameComp->setTooltip(getFile().getFullPathName());
    }

    void filenameComponentChanged(FilenameComponent*)
    {
      if (getFile() != filenameComp->getCurrentFile())
      {
        setFile(filenameComp->getCurrentFile());
        filenameComp->setTooltip(getFile().getFullPathName());
      }
    }

  private:
    FilenameComponent* filenameComp;
};

class FileChoice : public FilePropertyComponent, public PropertySetListener
{
  public:
    FileChoice(PropertySetSubject& properties,
               const String& name, const String& name_in_set,
               bool allow_edit,
               bool is_for_save,
               const String& wildcard = JUCE_T("*"))
        : FilePropertyComponent(name, false, allow_edit, is_for_save, wildcard),
          m_properties(properties),
          m_name_in_set(name_in_set)
    {
      setFile(properties.getValue(m_name_in_set));
      m_properties.addListener(*this);
    }

    virtual void onePropertyChanged(PropertySet& set)
    {
      m_file = File(m_properties.getValue(m_name_in_set));
      refresh();
    }
    
    virtual void setFile(const File& newFile)
    {
      if (m_file == newFile) return;
      m_file = newFile;
      m_properties.setValue(m_name_in_set, m_file.getFullPathName());
    }

    virtual const File getFile() const
    {
      return m_file;
    }

  private:
    PropertySetSubject& m_properties;
    File m_file;
    String m_name_in_set;
};

class AcPropertyChoice : public ChoicePropertyComponent, public PropertySetListener
{
  public:
    AcPropertyChoice(PropertySetSubject& properties,
                     const String& title_name,
                     const String& name_in_set)
        : ChoicePropertyComponent(title_name),
        m_index(0),
        m_properties(properties),
        m_name_in_set(name_in_set)
    {      
      properties.addListener(*this);
    }

    virtual void onePropertyChanged(PropertySet& set)
    {
      int new_index = choices.indexOf(m_properties.getValue(m_name_in_set));
      if (new_index >= 0)
        m_index = new_index;
      refresh();
    }
    
    virtual void setIndex(const int newIndex)
    {
      if (m_index == newIndex || newIndex < 0) return;
      m_index = newIndex;
      m_properties.setValue(m_name_in_set, getChoices()[m_index]);
    }

    virtual int getIndex() const
    {
      return m_index;
    }

  protected:
    int m_index;
    
  private:
    PropertySetSubject& m_properties;
    const String m_name_in_set;
};

class AcPropertyText : public TextPropertyComponent, public PropertySetListener
{
  public:
    AcPropertyText(PropertySetSubject& properties,
                   const String& title_name,
                   const String& name_in_set,
                   int max_num_char = 5)
        : TextPropertyComponent(title_name, max_num_char, 0),
        m_properties(properties),
        m_name_in_set(name_in_set)
    {
      m_text = properties.getValue(m_name_in_set);
      properties.addListener(*this);
    }

    virtual void onePropertyChanged(PropertySet& set)
    {
      m_text = m_properties.getValue(m_name_in_set);
      refresh();
    }
    
    virtual void setText(const String& s)
    {
      if (m_text == s) return;
      m_text = s;
      m_properties.setValue(m_name_in_set, m_text);
    }

    virtual const String getText() const
    {
      return m_text;
    }

  private:
    PropertySetSubject& m_properties;
    const String m_name_in_set;
    String m_text;
};

class AcPropertyBoolean : public BooleanPropertyComponent, public PropertySetListener
{
  public:
    AcPropertyBoolean(PropertySetSubject& properties,
                      const String& title_name,
                      const String& name_in_set)
        : BooleanPropertyComponent(title_name, JUCE_T("Yes"), JUCE_T("No")),
        m_properties(properties),
        m_name_in_set(name_in_set)
    {
      m_state = properties.getBoolValue(m_name_in_set);
      properties.addListener(*this);
    }

    virtual void onePropertyChanged(PropertySet& set)
    {
      m_state = m_properties.getBoolValue(m_name_in_set);
      refresh();
    }
    
    virtual void setState(const bool b)
    {
      m_state = b;
      m_properties.setValue(m_name_in_set, m_state);
      refresh();
    }

    virtual bool getState() const
    {
      return m_state;
    }

  private:
    PropertySetSubject& m_properties;
    const String m_name_in_set;
    bool m_state;
};

class SegmentorChoice : public AcPropertyChoice
{
  public:
    SegmentorChoice(PropertySetSubject& properties)
        : AcPropertyChoice(properties, "Segmentor", "segmentor")
    {
      choices.add("acsegmentor");
      choices.add("identical_pixels");
      choices.add("watershed");
      choices.add("egbis");
      int index = choices.indexOf(properties.getValue("segmentor"));
      if (index >= 0)
        m_index = index;
    }

};

class InitializerChoice : public AcPropertyChoice
{
  public:
    InitializerChoice(PropertySetSubject& properties)
        : AcPropertyChoice(properties, "Initializer", "initializer")
    {
      choices.add("identical_pixels");
      choices.add("watershed");
      choices.add("egbis");
      int index = choices.indexOf(properties.getValue("initializer"));
      if (index >= 0)
        m_index = index;
    }
};

#endif // ndef ACSEGMENTOR_gui_ac_properties_H
