//
// acsegmentor.cc
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//


#include "segmentor.h"
#include "segmentor_factory.h"
#include "ac_selector.h"
#include "ac_threshold_database.h"
#include "ac_learn_process.h"
#include "gui/ac_application.h"

#include <ntk/ntk_with_macros.h>

using namespace ntk;

void parse_command_line(int argc, char **argv, PropertySet* properties)
{
  ntk_assert(properties != 0, "Null pointer.");

  vul_arg<int> opt_debug_level("-d", "Debug level", 0);

  vul_arg<const char*> opt_image("--image", "Image to segment", 0);
  vul_arg<const char*> opt_output("--output", "Output image", "output.pgm");
  vul_arg<const char*> opt_output_edges("--output-edges", "Output edge image (optional)", 0);

  vul_arg<const char*>
      opt_initializer("--initializer", "Segmentor for initialization "
                                       "(identical_pixels, watershed, egbis).", "watershed");
  vul_arg<const char*>
      opt_segmentor("--segmentor", "Main segmentor "
                                   "(acsegmentor, identical_pixels, egbis, watershed).",
                    "acsegmentor");
  vul_arg<double>
      opt_watershed_sigma("--watershed-sigma", "Sigma for watershed", 0.8);
  vul_arg<double>
      opt_egbis_sigma("--egbis-sigma", "Sigma for egbis", 0.8);
  vul_arg<double>
      opt_egbis_k("--egbis-k", "K for egbis", 300);
  vul_arg<int>
      opt_egbis_min_size("--egbis-min-size", "Min size for egbis", 0);

  vul_arg<bool>
      opt_dont_enforce_meaningful_links("--dont-enforce-meaningful-links",
                                        "Do not ensure that all meaningful links will be preserved.",
                                        false);

  vul_arg<const char*>
      opt_selector("--selector", "Selector to use (acmvf, m)", "acmvf");
  vul_arg<bool> opt_no_mean("--no-mean", "Do not use mean in acmvf", false);
  vul_arg<bool> opt_no_dev("--no-deviation", "Do not use deviation in acmvf", false);
  vul_arg<bool> opt_no_frontier("--no-frontier", "Do not use frontier in acmvf", false);

  vul_arg<int>
      opt_max_regions("--maxregions", "Maximal number of regions (-1 = unlimited)", -1);

  vul_arg<const char*>
      opt_threshold_database("--threshold-database",
                             "File thresholds should be loaded from/saved to",
                             0);

  vul_arg<bool> opt_learning_mode("--learn", "Set learning mode.", 0);
  vul_arg<const char* > opt_learner("--learner", "Set learner (max,mean)", "mean");
  vul_arg<int> opt_width("--width", "Width for learning mode.", -1);
  vul_arg<int> opt_height("--height", "Height for learning mode.", -1);
  vul_arg<int> opt_nb_iterations("--nb-iterations", "Number of iterations in learning mode.", 1000);
  vul_arg<int> opt_nb_categories("--nb-categories", "Number of categories in learning mode.", 100);

  vul_arg_base::set_help_option("--help");
  vul_arg_parse(argc, argv);
  if (argc != 1)
  {
    vcl_cerr << "Command line error." << vcl_endl;
    vul_arg_base::display_usage_and_exit("Useless arguments on command line.");
  }

  ntk_debug_level = opt_debug_level();

  properties->setValue("initializer", String(opt_initializer()));
  properties->setValue("segmentor", String(opt_segmentor()));

  properties->setValue("watershed-sigma", opt_watershed_sigma());
  properties->setValue("egbis-sigma", opt_egbis_sigma());
  properties->setValue("egbis-k", opt_egbis_k());
  properties->setValue("egbis-min-size", opt_egbis_min_size());

  properties->setValue("keep_all_meaningful_links", !opt_dont_enforce_meaningful_links());

  properties->setValue("selector", String(opt_selector()));
  properties->setValue("use_mean", !opt_no_mean());
  properties->setValue("use_deviation", !opt_no_dev());
  properties->setValue("use_frontier", !opt_no_frontier());

  properties->setValue("maxregions", opt_max_regions());

  if (opt_threshold_database())
    properties->setValue("threshold_database", String(opt_threshold_database()));

  if (opt_learning_mode())
  {
    if (opt_image())
    {
      vcl_cerr << "Command line error: `--image' cannot be used in learning mode.\n"
               << "You can get some help with `acsegmentor --help'" << vcl_endl;
      exit(1);
    }
    if (opt_width() < 0) ntk_throw_exception("You must specify --width in learning mode.");
    if (opt_height() < 0) ntk_throw_exception("You must specify --height in learning mode.");
    properties->setValue(T("learning-mode"), true);
    properties->setValue(T("learning-width"), opt_width());
    properties->setValue(T("learning-height"), opt_height());
    properties->setValue(T("learning-learner"), String(opt_learner()));
    properties->setValue(T("learning-iterations"), opt_nb_iterations());
    properties->setValue(T("learning-categories"), opt_nb_categories());
  }
  else
  {
    if (!opt_image() || !opt_output())
    {
      vcl_cerr << "Command line error: `--image' and `--output' must be specified unless in learning mode.\n"
               << "You can get some help with `acsegmentor --help'"
               << vcl_endl;
      exit(1);
    }
    properties->setValue("image_to_segment", String(opt_image()));
    properties->setValue("output", String(opt_output()));
    if (opt_output_edges())
      properties->setValue("output-edges", String(opt_output_edges()));
  }
}

int main_nongui(const PropertySet& properties)
{
  SegmentorFactory factory;

  if (properties.getBoolValue("learning-mode", false))
  {
    AcLearnProcessPtr p = factory.createLearnProcess(properties);
    if (!p) ntk_throw_exception(T("Could not create learning process."));
    p->learn();
  }
  else // normal segmentation
  {
    SegmentorPtr segmentor = factory.createSegmentor(properties);
    if (!segmentor) ntk_throw_exception("Cannot create segmentor.");
    
    String image = properties.getValue("image_to_segment");
    vil_byte_image im = vil_load(image.toUTF8());
    if (!im) ntk_throw_exception("Could not load `" << image << "'");

    segmentor->segment(im);
    vil_image_view<uint8> labels;
    segmentor->getMeanImage(labels);
    vil_save(labels, properties.getValue("output"));

    if (properties.containsKey(T("output-edges")))
    {
      segmentor->getMeanImage(labels, true); // with borders.
      vil_save(labels, properties.getValue("output-edges"));
    }
  }

  if (properties.containsKey("threshold_database"))
  {
    ntk_assert(factory.acSelector(), "Threshold database only for AcSegmentors.");
    String f = properties.getValue("threshold_database");
    factory.acSelector()->thresholdDatabase()->saveToFile(f);
  }

  return 0;
}

#if JUCE_WIN32 // no console on windows

START_JUCE_APPLICATION(AcApplication)

#else // !JUCE_WIN32
int main(int argc, char** argv)
{
  if (argc <= 1 
      || String(argv[1]).startsWith(T("-psn_"))) // Ignore PSN on MacOSX
  {
    return JUCEApplication::main(argc, argv, new AcApplication());
  }

  initialiseJuce_NonGUI();
  ntk_log() << "Running in console mode.";

  PropertySet properties;
  parse_command_line(argc, argv, &properties);
  ntk_dbg_print(properties, 1);

  int return_value = 0;
  try
  {
    return_value = main_nongui(properties);
  }
  catch (const vcl_exception & e)
  {
    vcl_cerr << "Exception: " << vcl_endl << e.what() << vcl_endl;
    return_value = 1;
  }

  shutdownJuce_NonGUI();
  return return_value;
}
#endif // JUCE_WIN32
