//
// region.h
// 
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef ACSEGMENTOR_region_H
# define ACSEGMENTOR_region_H

# include <ntk/ntk.h>

# include <boost/multi_index/member.hpp>
# include <boost/multi_index/composite_key.hpp>

class Region
{
  public:
    typedef vcl_set<uint32>::const_iterator neighbors_const_iterator;
    typedef vcl_set<ntk::vil_point>::const_iterator points_const_iterator;
  
  public:
    Region() : m_id(-1) {}
    explicit Region(uint32 id) { m_id = id; }

  public:
    uint32 id() const { return m_id; }
    bool isValid() const { return m_id >= 0; }

    void mergeWith(const Region& rhs)
    {
      m_points.addArray(rhs.m_points);
      m_neighbors += rhs.m_neighbors;
      m_neighbors.erase(id());
    }
  
    uint32 size() const { return m_points.size(); }
    ntk::vil_point operator[] (uint32 index) const { return m_points[index]; }
    void addPoint(const ntk::vil_point& p) { m_points.add(p); }
  
    neighbors_const_iterator neighborsBegin() const { return m_neighbors.begin(); }
    neighbors_const_iterator neighborsEnd() const { return m_neighbors.end(); }
    uint32 nbNeighbors() const { return m_neighbors.size(); }
    void addNeighbor(uint32 r_id) { if (r_id != id()) m_neighbors.insert(r_id); }
    void removeNeighbor(uint32 id) { m_neighbors.erase(id); }

  private:
    int m_id;
    Array<ntk::vil_point> m_points;
    vcl_set<uint32> m_neighbors;
};

inline bool operator<(const Region& r1, const Region& r2)
{ return r1.id() < r2.id(); }

class RegionEdge
{
  public:
    RegionEdge(uint32 r1, uint32 r2)
    {
      if (r1 < r2)
      {
        m_lower_id = r1;
        m_upper_id = r2;
      }
      else
      {
        m_lower_id = r2;
        m_upper_id = r1;
      }
    }

    uint32 r1() const { return m_lower_id; }
    uint32 r2() const { return m_upper_id; }

  protected:
    uint32 m_lower_id;
    uint32 m_upper_id;
    
    friend struct RegionEdgeLabelKey;
    friend struct RegionEdgeMeaningfulThenScoreKey;
};

inline bool operator<(const RegionEdge& lhs, const RegionEdge& rhs)
{ 
  if (lhs.r1() == rhs.r1()) return lhs.r2() < rhs.r2();
  return lhs.r1() < rhs.r1();
}

/**
 * Unoriented edge between two regions. They are sorted by id.
 */
class RegionEdgeWithScore : public RegionEdge
{
  public:
    RegionEdgeWithScore(uint32 r1, uint32 r2, double score = 0, bool is_meaningful = false)
  : RegionEdge(r1, r2), m_score(score), m_is_meaningful(is_meaningful)
    {}

    void setScore(double s) { m_score = s; }
    double score() const { return m_score; }

    void setMeaningful(bool is_meaningful) { m_is_meaningful = is_meaningful; }
    bool isMeaningful() const { return m_is_meaningful; }

  private:
    double m_score;
    bool m_is_meaningful;
      
    friend struct RegionEdgeLabelKey;
    friend struct RegionEdgeMeaningfulThenScoreKey;
};

using namespace boost;
using namespace boost::multi_index;

// Do not use accessors because it implies a costly function call
// overhead: pointers on member functions cannot be inlined...

struct RegionEdgeLabelKey
{
  typedef composite_key<RegionEdgeWithScore,
            member<RegionEdge,uint32,&RegionEdgeWithScore::m_lower_id>,
            member<RegionEdge,uint32,&RegionEdgeWithScore::m_upper_id>
          > type;
};
  
struct RegionEdgeMeaningfulThenScoreKey
{
  typedef composite_key<RegionEdgeWithScore,
            member<RegionEdgeWithScore,bool,&RegionEdgeWithScore::m_is_meaningful>,
            member<RegionEdgeWithScore,double,&RegionEdgeWithScore::m_score>
          > type;
};

#endif // ndef ACSEGMENTOR_region_H
