//
// segmentor_factory.cc
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "segmentor_factory.h"
#include "segmentor.h"
#include "ac_segmentor.h"
#include "ac_explorer.h"
#include "ac_selector.h"
#include "ac_threshold_database.h"
#include "ac_learn_process.h"
#include "egbis.h"
#include "watershed.h"

#include <ntk/ntk_with_macros.h>
#include <vil/vil_convert.h>

void SegmentorFactory :: getDefaultProperties(PropertySet& properties)
{
#if JUCE_MAC
  File prefix("~/");
#else
  File prefix = File::getCurrentWorkingDirectory();
#endif
  properties.setValue("threshold_database", prefix.getChildFile("thresholds.xml").getFullPathName());

  properties.setValue("initializer", T("egbis"));
  properties.setValue("segmentor", T("acsegmentor"));

  properties.setValue("watershed-sigma", 0.8);
  properties.setValue("egbis-sigma", 0.8);
  properties.setValue("egbis-k", 300);
  properties.setValue("egbis-min-size", 0);
  properties.setValue("keep_all_meaningful_links", true);

  properties.setValue("selector", T("acmvf"));
  properties.setValue("use_mean", true);
  properties.setValue("use_deviation", true);
  properties.setValue("use_frontier", true);

  properties.setValue("maxregions", -1);
  
  // Learning
  properties.setValue(T("learning-width"), 256);
  properties.setValue(T("learning-height"), 256);
  properties.setValue(T("learning-learner"), T("mean"));
  properties.setValue(T("learning-iterations"), 1000);
  properties.setValue(T("learning-categories"), 100);
}

SegmentorPtr SegmentorFactory::
createEgbisSegmentor(const PropertySet& properties)
{
  double sigma = properties.getDoubleValue(T("egbis-sigma"), 0.5);
  double k = properties.getDoubleValue(T("egbis-k"), 500);
  int min_size = properties.getIntValue(T("egbis-min-size"), 0);
  return SegmentorPtr(new ExternalSegmentor(
                      new EgbisFunctor(sigma, k, min_size)));
}

SegmentorPtr SegmentorFactory::
createWatershedSegmentor(const PropertySet& properties)
{
  double sigma = properties.getDoubleValue(T("watershed-sigma"), 0.5);
  return SegmentorPtr(new ExternalSegmentor(
                      new WatershedFunctor(sigma)));
}

AcExplorerPtr SegmentorFactory::
createAcExplorer(const PropertySet& properties)
{
  String initializer_name = properties.getValue("initializer");
  SegmentorPtr initializer;

  if (initializer_name == T("identical_pixels"))
  {
    initializer.reset(new ExternalSegmentor(new IdenticalPixelsFunctor()));
  }
  else if (initializer_name == T("egbis"))
  {
    initializer = createEgbisSegmentor(properties);
  }
  else if (initializer_name == T("watershed"))
  {
    initializer = createWatershedSegmentor(properties);
  }

  if (!initializer) return AcExplorerPtr();

  {
    MergeBestExplorer* explorer = new MergeBestExplorer(initializer);
    if (properties.containsKey(T("keep_all_meaningful_links")))
      explorer->setKeepAllMeaningfulLinks(properties.getBoolValue(T("keep_all_meaningful_links")));
    if (properties.containsKey(T("maxregions")))
      explorer->setMaxNumOfRegions(properties.getIntValue(T("maxregions")));
    return AcExplorerPtr(explorer);
  }
}

AcSelectorPtr SegmentorFactory::
createAcSelector(const PropertySet& properties)
{
  AcThresholdDatabasePtr db (new AcThresholdDatabase());

  if (properties.getValue("selector") == String("acmvf"))
  {
    AcSelectorMVF* s = new AcSelectorMVF(db);
    s->setUseMean(properties.getBoolValue("use_mean"));
    s->setUseDeviation(properties.getBoolValue("use_deviation"));
    s->setUseFrontier(properties.getBoolValue("use_frontier"));
    return AcSelectorPtr(s);
  }

  if (properties.getValue("selector") == String("m"))
    return AcSelectorPtr(new AcSelectorMeanThreshold(db));

  return AcSelectorPtr();
}

AcSegmentorPtr SegmentorFactory::
createAcSegmentor(const PropertySet& properties)
{
  AcExplorerPtr explorer = createAcExplorer(properties);
  if (!explorer)
  {
    vcl_cerr << "Could not create explorer." << vcl_endl;
    return AcSegmentorPtr();
  }

  m_ac_selector = createAcSelector(properties);
  if (!m_ac_selector)
  {
    vcl_cerr << "Could not create selector." << vcl_endl;
    return AcSegmentorPtr();
  }

  m_ac_segmentor.reset(new AcSegmentor(explorer, m_ac_selector));

  if (properties.containsKey("threshold_database"))
  {
    File f (properties.getValue("threshold_database"));
    if (f.exists())
    {
      m_ac_selector->thresholdDatabase()->loadFromFile(f.getFullPathName());
    }
    else
    {
      m_ac_selector->thresholdDatabase()->saveToFile(f.getFullPathName());
    }
  }

  return m_ac_segmentor;
}

SegmentorPtr SegmentorFactory::
createSegmentor(const PropertySet& properties)
{
  clearAll();
  String segmentor = properties.getValue("segmentor");

  if (segmentor == T("identical_pixels"))
  {
    return SegmentorPtr(new ExternalSegmentor(new IdenticalPixelsFunctor()));
  }
  else if (segmentor == T("acsegmentor"))
  {
    return createAcSegmentor(properties);
  }
  else if (segmentor == T("egbis"))
  {
    return createEgbisSegmentor(properties);
  }
  else if (segmentor == T("watershed"))
  {
    return createWatershedSegmentor(properties);
  }

  return SegmentorPtr();
};

AcLearnProcessPtr SegmentorFactory ::
createLearnProcess(const PropertySet& properties)
{
  createSegmentor(properties);
  if (!acSegmentor()) ntk_throw_exception(T("Learning is only possible with acsegmentor."));
   
  AcLearnProcess* p = new AcLearnProcess(properties, acSegmentor());
  return AcLearnProcessPtr(p);
}
